"""Resource client for obtaining an access token.
"""
import json
import os
from collections import UserDict
from datetime import datetime as dt
from typing import Any, Dict, List

import iso8601
import requests

from ..client import Client
from .resource import ResourceClient


class TokenStorage(UserDict):
    """
    Dict based persistent storage for access tokens. Initializes the cache
    only with the valid (not expired) tokens and persists tokens on each
    call to create.
    """

    def __init__(self, tokens_file: str):
        self.tokens_file = tokens_file

        # if the tokens file is empty, it implies that the user has chosen
        # not to cache tokens, so this class just acts as an in memory dict
        if self.tokens_file == "":
            super().__init__()
            return

        try:
            with open(
                self.tokens_file, mode="r", encoding="utf-8"
            ) as tokens_fd:
                persisted_tokens = json.loads(tokens_fd.read())
                valid_tokens = {
                    token_id: persisted_tokens[token_id]
                    for token_id in persisted_tokens
                    if iso8601.parse_date(
                        persisted_tokens[token_id]["validUntil"]
                    ).timestamp()
                    > dt.now().timestamp()
                }

                super().__init__(valid_tokens)
        except FileNotFoundError:
            super().__init__()

    def __delitem__(self, key) -> None:
        super().__delitem__(key)
        self._write()

    def __setitem__(self, __key, __value) -> None:
        super().__setitem__(__key, __value)
        self._write()

    def _write(self):
        """write the tokens to persistent file"""

        if self.tokens_file == "":
            return

        tokens_path = os.path.dirname(self.tokens_file)
        # create directory if needed.
        try:
            os.mkdir(tokens_path)
        except OSError:
            pass  # ignore error creating directory.

        tokens_blob = json.dumps(self.data, sort_keys=True, indent=4)

        def opener(path, flags):
            return os.open(path, flags, 0o600)

        with open(
            self.tokens_file, mode="w", encoding="utf-8", opener=opener
        ) as tokens_file:
            print(tokens_blob, file=tokens_file)


class NoValidTokenException(Exception):
    """
    Exception for when there is no valid token in cache and you
    cannot retrieve/create more.
    """

    def __init__(self, *args: object) -> None:
        super().__init__("no valid token in cache", *args)


class AccessTokenClient(ResourceClient):
    """AccessTokenClient is used to obtain access tokens to
    authenticate to databases.
    """

    def __init__(
        self,
        cyral_client: Client,
        cache_tokens: bool = True,
        automatically_create: bool = True,
    ):
        super().__init__(cyral_client)
        tokens_file = ""
        if cache_tokens:
            tokens_file = os.path.join(
                "~", ".cyral", f".{cyral_client.cp_address}.tokens"
            )
            tokens_file = os.path.expanduser(tokens_file)
        self.token_cache = TokenStorage(tokens_file)
        self.automatically_create = automatically_create

    def _get_legacy(self) -> str:
        """
        retrieves the access token from the legacy route.
        """
        resp = self.do_get("/v1/opaqueToken/accessToken")

        # we check if the token has an ID because on the legacy route
        # there is no concept of token ID, so if the response has
        # an ID it means we are dealing with the new access token
        if "id" in resp:
            self.token_cache[resp["id"]] = resp
            return resp["accessToken"]
        return resp["accessToken"]

    def get(self, uuid: str = "") -> str:
        """
        retrieves a valid access token from the cache and prints
        its value. If no valid access token is cached, attempts
        to retrieve the token from the legacy token route. This
        will create a token automatically with the default
        parameters if there is a possibility of creating one.
        If it fails, raises NoValidTokenException.
        """
        try:
            return self.show(uuid)
        except NoValidTokenException:
            if self.automatically_create and not uuid:
                try:
                    token = self.create_token()
                    return token["accessToken"]
                except requests.exceptions.HTTPError as err:
                    if (
                        err.response is not None
                        and err.response.status_code
                        == requests.codes["not_found"]
                    ):
                        return self._get_legacy()
                    raise
        raise NoValidTokenException

    def show(self, uuid: str = "") -> str:
        """
        retrieves a valid access token from the cache.
        """

        if not uuid:
            for token in self.token_cache.data.values():
                return token["accessToken"]

        elif uuid in self.token_cache.data:
            return self.token_cache.data["accessToken"]

        raise NoValidTokenException

    def create_token(
        self,
        validity: int = 0,
        name: str = f"cyral-cli-{dt.now().strftime('%Y%m%d%H%M%S')}",
    ) -> Dict[str, Any]:
        """
        creates an access token with the given validity and name.
        The validity is in number of seconds.
        """
        data = {}
        if validity != 0:
            data["validity"] = f"{validity}s"
        if name != "":
            data["name"] = name

        token = self.do_post("/v1/accessTokens/tokens", data)
        self.token_cache[token["id"]] = token
        return token

    def list_tokens(self) -> List[Dict[str, Any]]:
        """
        lists the user's access tokens and informs whether they are
        cached or not.
        """
        tokens = self.do_get("/v1/accessTokens/tokens")["accessTokens"]
        for token in tokens:
            if token["id"] in self.token_cache.data:
                token["cached"] = True
        return tokens

    def delete_token(self, token_id: str):
        """
        deletes an access token
        """
        self.do_delete(f"/v1/accessTokens/tokens/{token_id}")
        if token_id in self.token_cache:
            del self.token_cache[token_id]



"""
Provides click options to add to commands.
"""
from typing import List

import click


# _command_help is a callback function to print the help for the current
# function and exit. We explicitly add a --help option to all groups and
# commands with this callback to avoid having the command handler
# execute when --help is specified.
def _command_help(ctx: click.Context, _, value):
    if value:
        print(ctx.command.get_help(ctx))
        ctx.exit()


# custom decorator to add a custom help option to commands.
def command_help_option():
    """
    decorator to add --help flag to command
    """
    return click.option(
        "--help",
        "-h",
        is_flag=True,
        show_default=False,
        help="Show help message and exit.",
        callback=_command_help,
        expose_value=False,
        is_eager=True,
    )


def output_format_option(formats: List[str], default_format: str):
    """
    decorator to add --output-format flag to command
    """
    return click.option(
        "--output-format",
        "-f",
        help="Output format",
        show_default=True,
        type=click.Choice(formats),
        show_choices=True,
        default=default_format,
    )


def token_cache_option():
    """
    decorator to add --use-token-cache flag to command
    """
    return click.option(
        "--use-token-cache/--no-use-token-cache",
        help="Whether to use local token cache.",
        type=click.BOOL,
        default=True,
        show_default=True,
    )


def token_option():
    """
    decorator to add --token flag to command
    """
    return click.option(
        "--token",
        help="Access token value to use (optional)",
        type=click.STRING,
        default="",
    )


def auto_token_option():
    """
    decorator to add --auto-token flag to command
    """
    return click.option(
        "--auto-generate-token/--no-auto-generate-token",
        "auto_token",
        type=click.BOOL,
        default=True,
        help="Generate a token automatically if needed",
        show_default=True,
    )


def token_id_option(required: bool = True):
    """
    decorator to add --token-id flag to command
    """

    def callback(_, param: click.Option, value: str):
        if required and not value:
            raise click.BadParameter(f"{param.name} must not be empty")
        return value

    return click.option(
        "--token-id",
        help="ID of the token on which the action will be performed",
        type=click.STRING,
        default="",
        callback=callback,
    )



"""
Cyral CLI
"""
from .commands import cli


def main():
    """entrypoint for the Cyral CLI"""
    # pylint: disable=no-value-for-parameter
    cli(obj={}, auto_envvar_prefix="CYRAL")


if __name__ == "__main__":
    main()



"""Configure AWS profile to access S3 via Cyral sidecar"""

import configparser
import importlib.resources
import os
import subprocess  # nosec: B404
from pathlib import Path

from botocore.configloader import raw_config_parse
from botocore.exceptions import ConfigNotFound
from botocore.session import get_session

S3_PROXY_PLUGIN = "awscli-plugin-proxy"
S3_PROXY_PACKAGE = S3_PROXY_PLUGIN.replace("-", "_")


class AWSConfig:
    """AWSConfig class provides access to the AWS config file"""

    def __init__(self, config_file: str) -> None:
        self.config_file = config_file
        try:
            self.config = raw_config_parse(self.config_file)
        except ConfigNotFound:
            self.config = {}

    def get_profile(self, profile_name: str) -> dict:
        """Get the profile with the given name."""
        return self.config.get(f"profile {profile_name}", {})

    def set_profile(self, profile_name: str, profile: dict) -> None:
        """Set the profile with the given name."""
        self.config[f"profile {profile_name}"] = profile

    def get_section(self, section_name: str) -> dict:
        """Get the section with the given name."""
        return self.config.get(section_name, {})

    def set_section(self, section_name: str, section: dict) -> None:
        """Set the section with the given name."""
        self.config[section_name] = section

    def write(self) -> None:
        """Write the config to the file"""
        config_parser = configparser.ConfigParser(interpolation=None)
        for (section, kvs) in self.config.items():
            config_parser.add_section(section)
            for (key, value) in kvs.items():
                if isinstance(value, dict):
                    val = ""
                    for (sub_key, sub_value) in value.items():
                        val += f"\n{sub_key} = {sub_value}"
                else:
                    val = value
                config_parser.set(section, key, val)
        with open(self.config_file, "w", encoding="utf-8") as config_file:
            config_parser.write(config_file)


class AWSCredentials(AWSConfig):
    """AWSCredentials class provides access to the AWS credentials file"""

    def get_profile(self, profile_name: str) -> dict:
        """
        Get the profile with the given name. This overrides the base class
        method because the names of profile sections in the credentials file
        do not have the "profile" prefix.
        """
        return self.config.get(profile_name, {})

    def set_profile(self, profile_name: str, profile: dict) -> None:
        """
        Set the profile with the given name. This overrides the base class
        method because the names of profile sections in the credentials file
        do not have the "profile" prefix.
        """
        self.config[profile_name] = profile


def _get_config_file_path(file: str) -> str:
    # we use the botocore and awscli existing code to get this done.
    session = get_session()
    config_path = session.get_config_variable(file)
    config_path = os.path.expanduser(config_path)
    return config_path


def update_aws_creds(
    access_token: str,
    user_email: str,
    aws_profile_name: str,
    silent: bool,
    user_account: str,
) -> None:
    """Update AWS credentials based on the Cyral access token."""
    creds = AWSCredentials(_get_config_file_path("credentials_file"))
    profile = creds.get_profile(aws_profile_name)

    key_id = f"{user_email}:{access_token}"
    if user_account != "":
        key_id += f":{user_account}"
    values = {
        "aws_access_key_id": key_id,
        "aws_secret_access_key": "none",
    }
    profile.update(values)
    creds.set_profile(aws_profile_name, profile)
    creds.write()

    if not silent:
        print(f"Updated S3 token for AWS profile '{aws_profile_name}'")
        if aws_profile_name != "default":
            print(
                "\nTo use this profile, specify the profile name using "
                "--profile, as shown:\n\n"
                f"aws s3 ls --profile {aws_profile_name}\n"
            )


class S3ProxyPluginNotInstalled(Exception):
    """S3ProxyPluginNotInstalled exception is raised if the user does
    not have the S3 proxy plugin installed."""


def configure_s3_proxy_settings(
    aws_profile_name: str,
    sidecar_endpoint: str,
    ca_bundle: str,
) -> None:
    """Configure S3 proxy settings in the AWS profile."""
    if not _s3_plugin_is_installed():
        raise S3ProxyPluginNotInstalled(
            "Please first install S3 proxy plugin using the command:\n\n"
            + f"pip3 install {S3_PROXY_PLUGIN}"
        )
    conf = AWSConfig(_get_config_file_path("config_file"))
    try:
        _update_s3_proxy_plugins(conf)
        _update_ca_bundle(conf, aws_profile_name, ca_bundle)
        conf.write()
        update_s3_proxy_endpoint(aws_profile_name, sidecar_endpoint)
    except Exception as ex:
        # pylint: disable=broad-exception-raised
        raise Exception("error configuring S3 proxy settings") from ex


def _save_ca_bundle(ca_bundle: str, direname: str) -> str:
    ca_bundle_default_path = Path(direname) / "cyral_ca_bundle.pem"
    with open(ca_bundle_default_path, "w", encoding="utf-8") as file:
        file.write(ca_bundle)
    return str(ca_bundle_default_path)


def _update_ca_bundle(
    conf: AWSConfig,
    aws_profile_name: str,
    ca_bundle: str,
) -> None:
    config_path = _get_config_file_path("config_file")
    ca_bundle_direname = os.path.dirname(config_path)
    cyral_ca_bundle_file = _save_ca_bundle(ca_bundle, ca_bundle_direname)
    profile = conf.get_profile(aws_profile_name)
    profile.update({"ca_bundle": cyral_ca_bundle_file})
    conf.set_profile(aws_profile_name, profile)


def _update_s3_proxy_plugins(conf: AWSConfig) -> None:
    installed_plugin_name = S3_PROXY_PACKAGE
    plugins = conf.get_section("plugins")
    plugins.update({"s3-proxy": installed_plugin_name})
    if _get_cli_version() == "v2":
        plugins.update(
            {"cli_legacy_plugin_path": _get_cli_legacy_plugin_path()},
        )
    conf.set_section("plugins", plugins)


def update_s3_proxy_endpoint(aws_profile_name: str, endpoint: str) -> None:
    """update the S3 proxy endpoint"""
    endpoint = endpoint.replace("http://", "").replace("https://", "")
    conf = AWSConfig(_get_config_file_path("config_file"))

    profile = conf.get_profile(aws_profile_name)
    for command in ["s3", "s3api"]:
        value = profile.get(command, {})
        value.update({"proxy": f"http://{endpoint}"})
        profile.update({command: value})
    conf.write()


def s3_proxy_is_configured(aws_profile_name: str) -> bool:
    """Check if S3 proxy is setup in the specified AWS profile."""
    conf = AWSConfig(_get_config_file_path("config_file"))
    plugins = conf.get_section("plugins")
    profile = conf.get_profile(aws_profile_name)
    # a correctly configured config looks like the following:
    # {
    #     "plugins":{
    #         "s3-proxy":"awscli_plugin_s3_proxy"
    #     },
    #     "profiles":{
    #         "cyral":{
    #             "ca_bundle":"/home/user/.aws/cyral_ca_bundle.pem",
    #             "s3":{
    #                 "proxy":"http://sidecar.endpoint:453"
    #             },
    #             "s3api":{
    #                 "proxy":"http://sidecar.endpoint:453"
    #             }
    #         }
    #     }
    # }
    if (  # pylint: disable=too-many-boolean-expressions
        plugins.get("s3-proxy")
        and profile.get("ca_bundle")
        and Path(profile["ca_bundle"]).is_file()
        and profile.get("s3")
        and profile["s3"].get("proxy")
        and profile.get("s3api")
        and profile["s3api"].get("proxy")
        and _s3_plugin_is_installed()
    ):
        return True
    return False


def _get_cli_legacy_plugin_path() -> str:
    # should be the dir of the installed S3_PROXY_PLUGIN
    try:
        pkg = importlib.resources.files(S3_PROXY_PACKAGE)
        return os.path.dirname(str(pkg))
    except ModuleNotFoundError as ex:
        raise Exception(  # pylint: disable=broad-exception-raised
            "Failed to find a legacy plugin path for AWS cli.",
        ) from ex


def _s3_plugin_is_installed() -> bool:
    try:
        _ = importlib.resources.files(S3_PROXY_PACKAGE)
        return True
    except ModuleNotFoundError:
        return False


def _get_cli_version() -> str:
    # returns the major version
    try:
        cli_output = subprocess.check_output(  # nosec: B603 B607
            ["aws", "--version"],
        ).decode("utf-8")
        if not cli_output.startswith("aws-cli"):
            raise AssertionError(f"unrecognized AWS CLI version: {cli_output}")
        # example output: aws-cli/2.1.15 Python/3.7.3 ...
        aws_version = cli_output.split("/")[1]
        major_version = "v" + aws_version[0]
        return major_version
    except (subprocess.CalledProcessError, AssertionError) as ex:
        raise Exception(  # pylint: disable=broad-exception-raised
            "Failed to get AWS cli version. Make sure AWS CLI is installed!",
        ) from ex
